"""
The `processor.py` module provides functionality to process and save articles and their comments
from Spiegel given a fetched archive.
"""


import asyncio
from asyncio import AbstractEventLoop
import json
import logging
import os
import time
from concurrent.futures import ThreadPoolExecutor
from datetime import date
from pathlib import Path

from spon_api.article import Article
from spon_api.comments import Comments


class Processor:
    """
    A class used to process and save Spiegel articles and comments based on specified criteria.

    Attributes
    ----------
    exclusion_count : int
        Counter for the number of articles excluded based on criteria.
    exclusions : list[str]
        A list of URL patterns to exclude from processing.
    targets : list[str]
        A list of target types to process ('articles' or 'comments').
    output_path : Path
        The path where the processed data will be saved.
    executor : ThreadPoolExecutor
        Executor to run blocking tasks in separate threads.

    Methods
    -------
    process_archive(archive: list[dict], target_date: date, chunk_size: int)
        Asynchronously process the article archive in chunks.
    process_articles_chunk(chunk: list[dict], target_date: date)
        Asynchronously process a chunk of articles.
    _process_article(url: str, index: int, target_date: date)
        Processes a single article and corresponding comments and saves the data.
    _get_article_by_url(url: str) -> dict
        A static method to fetch and parse the article data by its URL.
    _get_comments_by_article_id(article_id: str) -> list[dict]
        A static method to fetch and parse the comments data by article ID.
    """

    def __init__(
        self,
        exclusion_count: int,
        exclusions: list[str],
        targets: list[str],
        output_path: Path,
        executor: ThreadPoolExecutor,
    ) -> None:
        """
        Constructs the Processor object.

        Parameters
        ----------
        exclusion_count : int
            Counter for the number of articles excluded based on criteria.
        exclusions : list[str]
            A list of URL patterns to exclude from processing.
        targets : list[str]
            A list of target types to process ('articles' or 'comments').
        output_path : Path
            The path where the processed data will be saved.
        executor : ThreadPoolExecutor
            Executor to run blocking tasks in separate threads.
        """
        self.exclusion_count: int = exclusion_count
        self.exclusions: list[str] = exclusions
        self.targets: list[str] = targets
        self.output_path: Path = output_path
        self.executor: ThreadPoolExecutor = executor

    async def process_archive(
        self, archive: list[dict], target_date: date, chunk_size: int
    ) -> None:
        """
        Asynchronously process the archive data in chunks.
        This method divides the archive data into chunks and processes each chunk concurrently.

        Parameters
        ----------
        archive : list[dict]
            The archive data to be processed.
        target_date : date
            The date for which the archive data is being processed.
        chunk_size : int
            The number of articles to process concurrently in a single batch.

        Returns
        -------
        None
        """
        tasks: list = []

        for archive_entry in range(0, len(archive), chunk_size):
            chunk: list[dict] = archive[archive_entry : archive_entry + chunk_size]
            tasks.append(self.process_articles_chunk(chunk, target_date))

        await asyncio.gather(*tasks)
        logging.info("Date finished")

    async def process_articles_chunk(
        self, chunk: list[dict], target_date: date
    ) -> None:
        """
        Asynchronously process a chunk of articles.
        This method processes articles in the given chunk and filters them based on exclusions.

        Parameters
        ----------
        chunk : list[dict]
            The chunk of articles to be processed.
        target_date : date
            The date for which the articles are being processed.

        Returns
        -------
        None
        """
        loop: AbstractEventLoop = asyncio.get_event_loop()
        tasks: list = []

        for index, article in enumerate(chunk):
            if any(exclusion in article["url"] for exclusion in self.exclusions):
                self.exclusion_count += 1
            else:
                tasks.append(
                    loop.create_task(
                        self._process_article(article["url"], index, target_date)
                    )
                )
        await asyncio.gather(*tasks)

    async def _process_article(self, url: str, index: int, target_date: date) -> None:
        """
        Processes a single article and saves the data.
        This method fetches article and comment data then saves them in the designated output directory.

        Parameters
        ----------
        url : str
            The URL of the article to be processed.
        index : int
            The index of the article within the current processing batch.
        target_date : date
            The date for which the article is being processed.

        Returns
        -------
        None
        """
        loop: AbstractEventLoop = asyncio.get_event_loop()

        if any(exclusion in url for exclusion in self.exclusions):
            self.exclusion_count += 1
            return

        try:
            article: dict = await loop.run_in_executor(
                self.executor, self._get_article_by_url, url
            )
            comments: list[dict] = await loop.run_in_executor(
                self.executor, self._get_comments_by_article_id, article["id"]
            )

            for target in self.targets:
                if target == "comments":
                    target_data: list[dict] = comments
                    output_dir = (
                        self.output_path
                        / f"comments/{target_date.year}/{target_date.month}"
                    )
                elif target == "articles":
                    target_data: dict = article
                    output_dir: Path = (
                        self.output_path
                        / f"articles/{target_date.year}/{target_date.month}"
                    )
                else:
                    logging.error(f"Unknown target type: {target}")
                    continue

                os.makedirs(output_dir, exist_ok=True)
                output_file: Path = (
                    output_dir
                    / f"{target_date.year}-{target_date.month}-{target_date.day}-{target}{index}.json"
                )
                with output_file.open("w") as f:
                    json.dump(target_data, f)

        except Exception as e:
            logging.error(f"Error processing article {url}: {str(e)}")
            time.sleep(60)

    @staticmethod
    def _get_article_by_url(url: str) -> dict:
        """
        Fetch and parse the article data by its URL.

        Parameters
        ----------
        url : str
            The URL of the article to be fetched and parsed.

        Returns
        -------
        dict
            The parsed article data.
        """
        article_object: Article = Article(url)
        article_object.fetch()
        article_data: dict = article_object.parse()
        return article_data

    @staticmethod
    def _get_comments_by_article_id(article_id: str) -> list[dict]:
        """
        Fetch and parse the comments data by article ID.

        Parameters
        ----------
        article_id : str
            The ID of the article whose comments need to be fetched and parsed.

        Returns
        -------
        list[dict]
            The parsed comments data.
        """
        comment_object: Comments = Comments(article_id)
        comment_object.fetch()
        comment_data: list[dict] = comment_object.parse()
        return comment_data
