"""
The `spiegel_archive_fetcher.py` module provides functionality to fetch and parse archive
from Spiegel for a specific date giving information on all articles on that date.

"""

import asyncio
from asyncio import AbstractEventLoop
from concurrent.futures import ThreadPoolExecutor
from datetime import date
from typing import Any

from spon_api.archive import Archive


class SpiegelArchiveFetcher:
    """
    A class used to fetch and parse archived Spiegel articles for a given date.

    Attributes
    ----------
    executor : ThreadPoolExecutor
        Executor to run blocking tasks in separate threads.

    Methods
    -------
    fetch_archive(target_date: date) -> list[dict[str, Any]]
        Asynchronous wrapper to fetch and parse archive data for a given date.
    _fetch_and_parse_archive(target_date: date) -> list[dict[str, Any]]
        A static method to fetch and parse the archive data for a given date.
    """

    def __init__(self, executor: ThreadPoolExecutor) -> None:
        """
        Constructs the SpiegelArchiveFetcher object.

        Parameters
        ----------
        executor : ThreadPoolExecutor
            Executor to run blocking tasks in separate threads.
        """
        self.executor: ThreadPoolExecutor = executor

    async def fetch_archive(self, target_date: date) -> list[dict[str, Any]]:
        """
        Asynchronously fetch and parse archive data for a given date.
        This method runs the archive fetching and parsing operation in an executor
        and returns the parsed data.

        Parameters
        ----------
        target_date : date
            The date for which the archive data should be fetched.

        Returns
        -------
        list[dict[str, Any]]
            A list of dictionaries containing the parsed archive data. Each entry is an article.
        """
        loop: AbstractEventLoop = asyncio.get_event_loop()
        archive: list[dict] = await loop.run_in_executor(
            self.executor,
            self._fetch_and_parse_archive,
            target_date,
        )
        return archive

    @staticmethod
    def _fetch_and_parse_archive(target_date: date) -> list[dict[str, Any]]:
        """
        Fetch and parse the archive data for a given date.
        This method uses the Archive class to fetch the HTML data for the target date
        and then parse the data to extract article details.

        Parameters
        ----------
        target_date : date
            The date for which the archive data should be fetched.

        Returns
        -------
        list[dict[str, Any]]
            A list of dictionaries containing the parsed archive data. One article per entry.
        """
        archive: Archive = Archive(target_date)
        archive.fetch()
        archive_data: list[dict[str, Any]] = archive.parse()
        return archive_data
